#! /usr/bin/env python

# Copyright (C) 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Script for generating Sphinx-friendly help for plugins.

Usage is:

    document-plugins.py xxx

where xxx is the name of an installed plugin. This will generate a
file called xxx-plugin.txt in the current directory.
"""


import os, re

from bzrlib import (
    _format_version_tuple,
    commands as mod_commands,
    errors,
    option,
    plugin as mod_plugin,
    initialize,
    )

_REGISTRY_NAME = "plugins-registry.ini"


def build_help_for_plugin(plugin, outf=None):
    """Build the help for a plugin.

    :param plugin: plugin object
    :param outf: output stream. If None, one is generated based
      on the plugin name.
    """
    if os.path.exists(_REGISTRY_NAME):
        reg_file = _REGISTRY_NAME
        out_dir = "plugins/en"
    elif os.path.exists("../../%s" % _REGISTRY_NAME):
        reg_file = "../../%s" % _REGISTRY_NAME
        out_dir = "."
    else:
        raise errors.BzrError("%s not found in . or ../.." % _REGISTRY_NAME)
    if outf is None:
        filename = "%s/%s-plugin.txt" % (out_dir,
            plugin.name.replace('_', '-'))
        print "creating %s ..." % (filename,)
        outf = open(filename, 'w')
    outf.write(get_plugin_help_text_for_sphinx(plugin, reg_file))
    for cmd_object in get_commands_for_plugin(plugin):
        outf.write(get_command_help_text_for_sphinx(cmd_object, level2=None,
            include_from=False, see_also_prefix=None))
    if outf and outf != sys.stdout:
        outf.close


_plugin_help_index = None
_plugin_registry = None


def get_plugin_info(name, registry_filename):
    """Return information from the plugin registry as a dictionary.
    
    :param name: name is nice format, e.g. foo-bar
    """
    global _plugin_registry
    if _plugin_registry is None:
        import ConfigParser
        _plugin_registry = ConfigParser.ConfigParser()
        _plugin_registry.read([registry_filename])
    section = "plugin %s" % name
    return dict(_plugin_registry.items(section))


def get_plugin_help_text_for_sphinx(plugin, registry_filename, level1='=',
        level2='-'):
    global _plugin_help_index
    if _plugin_help_index is None:
        _plugin_help_index = mod_plugin.PluginsHelpIndex()

    # Format the title
    nice_name = plugin.name.replace('_', '-')
    plugin_info = get_plugin_info(nice_name, registry_filename)
    title = plugin.name
    purpose = plugin_info['purpose']
    title = "``%s`` - %s" % (nice_name, purpose)

    # Format the overview
    overview = ''
    topics = _plugin_help_index.get_topics(plugin.name)
    if topics:
        overview = _format_overview_text(topics[0].get_help_text())
        #title = "%s - %s" % (plugin.name, lines[0])

    # Find the version information
    version_info = plugin.version_info()
    if version_info is None:
        version = "Not specified"
    else:
        try:
            version = _format_version_tuple(version_info)
        except ValueError:
            # Version info fails the expected rules
            version = "%s" % (version_info,)
        except IndexError:
            # Version info fails the expected rules
            version = "%s" % (version_info,)
    header_attrs = [('Version', version)]

    # Add the other header attributes from the registry
    branch_info = plugin_info.get('link.branch', 'Not specified')
    home_info = plugin_info.get('link.home')
    if branch_info.startswith('lp:'):
        proj = branch_info[3:]
        branch_info = "`%s <https://code.launchpad.net/%s>`_" % \
            (branch_info, proj)
        if home_info is None:
            home_info = "`<https://launchpad.net/%s>`_" % proj
    header_attrs.append(('Branch', branch_info))
    devel_info = plugin_info.get('link.development')
    if devel_info:
        header_attrs.append(('Development', devel_info))
    if home_info:
        header_attrs.append(('Home page', home_info))
    download_info = plugin_info.get('link.download')
    if download_info:
        header_attrs.append(('Downloads', download_info))
    owner = plugin_info.get('owner')
    if owner:
        owner_info = "`%s <https://launchpad.net/~%s>`_" % (owner, owner)
        header_attrs.append(('Owner', owner_info))
    header_attrs.append(('GNU/Linux', _format_platform(
        plugin_info.get('platform.gnulinux'))))
    header_attrs.append(('Windows', _format_platform(
        plugin_info.get('platform.windows'))))
    header_attrs.append(('Mac OS X', _format_platform(
        plugin_info.get('platform.mac-osx'))))

    # Format it
    result = format_section(title, '', level1)
    result += _format_attributes(header_attrs)
    result += overview
    result += "\n\n"
    return result


def _format_platform(p):
    if p == 'True':
        return 'Yes'
    elif p == 'False':
        # Should we be more explicit here with 'Not applicable' say?
        return 'No'
    else:
        return 'Maybe (please test)'


def _format_attributes(nv_pairs):
    header = "%s %s" % ('=' * 16, '=' * 60)
    lines = [header]
    for n, v in nv_pairs:
        lines.append("%-16s %-60s" % (n, v))
    lines.append(header)
    return "\n".join(lines) + "\n\n"


def _format_overview_text(text):
    """Strip out existing heading underlines."""
    lines = text.split('\n')
    result = []
    for line in lines:
        if re.search('^=+$', line):
            # Bold the previous line instead
            if result:
                result[-1] = "**%s**" % (result[-1])
            continue
        result.append(line)
    return '\n'.join(result)


# cache of commands by plugin
_plugin_commands = None
def get_commands_for_plugin(plugin):
    """Return a list of command sorted by name for a plugin."""
    plugin_name = plugin.name
    global _plugin_commands
    if _plugin_commands is None:
        # Initialise the cache
        names = list(mod_commands.all_command_names())
        commands = ((n, mod_commands.get_cmd_object(n)) for n in names)
        shown_commands = [(n, o) for n, o in commands if not o.hidden]
        _plugin_commands = {}
        for cmd_name, cmd_object in shown_commands:
            cmd_plugin_name = cmd_object.plugin_name()
            if cmd_plugin_name is not None:
                _plugin_commands.setdefault(cmd_plugin_name, []).append(cmd_object)
    return sorted(_plugin_commands.get(plugin_name, []))
 

def get_command_help_text_for_sphinx(cmd, level1='-', level2='~',
        include_from=True, see_also_prefix=':doc:'):
    result = format_section(cmd.name(), '', level1)
    doc = cmd.help()
    if doc is None:
        return result

    # Extract the summary (purpose) and sections out from the text.
    # If a custom usage section was provided, use it instead.
    purpose, sections, order = cmd._get_help_parts(doc)
    if sections.has_key('Usage'):
        usage = sections.pop('Usage')
    else:
        usage = cmd._usage()
    result += format_section("Purpose", purpose, level2)
    result += format_section("Usage", usage, level2)

    # Add the options
    #
    # XXX: optparse implicitly rewraps the help, and not always perfectly,
    # so we get <https://bugs.launchpad.net/bzr/+bug/249908>.  -- mbp
    # 20090319
    options = option.get_optparser(cmd.options()).format_option_help()
    if options.startswith('Options:'):
        options_text = options[len('Options:'):]
    elif options.startswith('options:'):
        # Python 2.4 version of optparse
        options_text = options[len('options:'):]
    else:
        # When does this case happen?
        options_text = options
    result += format_section("Options", deindent(options_text), level2)

    # Add the description
    if sections.has_key(None):
        text = sections.pop(None)
        text = '\n'.join(text.splitlines())
        result += format_section('Description', text, level2)

    # Add the custom sections (e.g. Examples).
    # XXX: These are indented already in the source by a space or two
    # so we may to deindent them (after checking that) in the future.
    if sections:
        for label in order:
            if sections.has_key(label):
                text = sections[label]
                result += format_section(label, deindent(text), level2)
        result += '\n'

    # Add the aliases, source (plug-in) and see also links, if any
    if cmd.aliases:
        result += format_section('Aliases', ', '.join(cmd.aliases), level2)
    if include_from:
        plugin_name = cmd.plugin_name()
        if plugin_name is not None:
            result += format_section('From', 'plugin "%s"' % plugin_name, level2)
    see_also = cmd.get_see_also()
    if see_also:
        if see_also_prefix:
            see_also_links = ["%s`%s`" % (see_also_prefix, topic)
                for topic in see_also]
        else:
            see_also_links = ["%s" % (topic,) for topic in see_also]
        result += format_section('See also', ', '.join(see_also_links), level2)
    return result


def format_section(hdg, text, hdg_underline):
    """Format a section given a heading and text content."""
    if hdg_underline is None:
        # Make it bold instead of a heading
        result = "**%s**\n\n" % (hdg,)
    else:
        result = "%s\n%s\n\n" % (hdg, hdg_underline * len(hdg))
    if text:
        result += "%s\n\n" % (text,)
    return result


def deindent(text):
    """Remove common leading spaces off lines in text."""
    lines = text.split("\n")
    indents = []
    for line in lines:
        if len(line) > 0:
            indents.append(len(line) - len(line.lstrip(' ')))
    indent = min(indents)
    if indent:
        lines = [l[indent:] for l in lines]
    return '\n'.join(lines)


_showed_plugins = False

def _show_installed_plugins(plugins):
    global _showed_plugins
    if not _showed_plugins:
        nice_names = [p.replace('_', '-') for p in plugins]
        print "installed plugins are %s" % (sorted(nice_names),)
        _showed_plugins = True


# Useful sets to build
ALIASES = {
    'ALL': """
        launchpad loggerhead webdav xmloutput
        explorer extmerge difftools gtk qbzr vimdiff
        automirror keywords push_and_update repo_push upload
        cia email
        checkeol pqm testrunner text_checker
        loom pipeline rebase undelete
        builder externals scmproj merge_into
        bisect diffstat search stats
        automv bookmarks bzrtools
        git hg svn
        bazimport cvsps_import fastimport
        dns_cache historycache service
        ping repodetails
        docdiff x_bit
        builddeb dbus pager
""".strip().split(),
}

if __name__ == '__main__':
    import sys

    interesting = sys.argv[1:]
    if len(interesting) == 0:
        print "usage document-plugin.py plugin-name"
        sys.exit(1)
    outf = None
    
    initialize()

    mod_plugin.load_plugins()
    mod_commands.install_bzr_command_hooks()
    plugins = mod_plugin.plugins()
    if interesting[0] in ALIASES:
        interesting = ALIASES[interesting[0]]
    for plugin_name in interesting:
        internal_name = plugin_name.replace('-', '_')
        try:
            plugin_object = plugins[internal_name]
        except KeyError:
            _show_installed_plugins(plugins.keys())
            print "WARNING: plugin %s is not installed - skipping." % plugin_name
        else:
            build_help_for_plugin(plugin_object, outf)
